import json
from typing import List, Dict, Any, Optional
from anc.data.anc_processor import Processor   

class MultiDialogProcessor(Processor):
    def __init__(self):
        super().__init__()
    
    def transform(self, item: List[Dict[str, str]], is_last_sample: bool = False) -> List[Dict[str, Any]]:
        processed_items = []
        # Extract system message if present
        system_message = None
        for msg in item:
            if msg['role'] == 'system':
                system_message = msg['content']
                break
        
        if system_message is None:
            system_message = ''
        
        # Build dialogue history for all turns
        dialogue = []
        for msg in item:
            if msg['role'] in ['user', 'assistant']:
                dialogue.append({
                    'role': msg['role'],
                    'content': msg['content']
                })
        
        # Process each turn in the conversation
        for i in range(len(dialogue)):
            # Only process user messages
            if i % 2 != 0 or dialogue[i]['role'] != 'user':
                continue
                
            # Check if there's a matching assistant response
            if i + 1 >= len(dialogue) or dialogue[i+1]['role'] != 'assistant':
                continue
                
            # Get history up to this point (including current user message)
            history = dialogue[:i+1]
            
            # Format for OpenAI API
            messages = []
            if system_message:
                messages.append({"role": "system", "content": system_message})
            messages.extend(history)
            
            # Get expected assistant response for output length calculation
            assistant_response = dialogue[i+1]['content']
            # Estimate token length - approximate calculation
            output_len = len(assistant_response.split()) * 2
            
            # Create a prompt in the format expected by vllm OpenAI API
            prompt = {
                "messages": messages
            }
            print(f'output_len: {output_len}, total_turns: {len(dialogue) // 2}')
            processed_item = {
                'prompt': json.dumps(prompt),  # OpenAI API format
                'output_len': output_len,
                'conversation_id': hash(str(item)),  # Unique ID for this conversation
                'turn_index': i // 2,  # Turn number in this conversation
                'total_turns': len(dialogue) // 2  # Total turns in conversation
            }
            processed_items.append(processed_item)
 
        return processed_items
    
    def batch_transform(self, list_of_items: List[List[Dict[str, str]]], is_last_batch: bool = False) -> List[Dict[str, Any]]:
        batch_processed_items = []
        for item in list_of_items:
            batch_processed_items.extend(self.transform(item))
        return batch_processed_items



import random
from anc.data.anc_sampler import AncSampler


class SimpleDataset:
    def __init__(self, sub_ds_count, sub_ds_size):
        self.sub_ds_count = sub_ds_count
        self.sub_ds_size = sub_ds_size

    def get_sub_lengths(self):
        return [self.sub_ds_size] * self.sub_ds_count

    def __len__(self):
        return self.sub_ds_count * self.sub_ds_size


def check_sampler_indices(shuffle, batch_size, resume_step):
    # create a fake dataset with 3 files and each file contains 100 elements
    sub_ds_size = 40
    sub_ds_count = 3
    ds = SimpleDataset(sub_ds_count, sub_ds_size)
    world = 4
    rank = 0
    num_workers = 2
    drop_last = False
    num_per_rank = (sub_ds_size * sub_ds_count + (world - 1) * int(not drop_last)) // world
    # with the above parameters (if shuffle = False), the data idxs that rank 0 would generate are [0, 30), totally 30 elements
    # since batch size = 4, the batch number would be 8, with last batch only containing 2 elements
    # thus worker 0 of rank 0 would generate [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    #      worker 1 of rank 0 would generate [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29]
    sampler = AncSampler(ds, batch_size, world, rank, num_workers, shuffle=False, drop_last=drop_last, resume_step=resume_step)
    assert len(sampler) == (num_per_rank + batch_size - 1) // batch_size
    indices = []
    for item in sampler:
        cur_index, is_last_batch = item
        if not is_last_batch:
            # cur_index would be [] if resume_step % num_workers != 0
            assert len(cur_index) == batch_size or len(cur_index) == 0
        if is_last_batch:
            # each worker's last batch would set is_last_batch to True
            # so we have num_workers is_last_batch
            assert sampler.step >= len(sampler) - num_workers
        indices += cur_index
    if not shuffle and resume_step == 0:
        assert indices[:16] == [0, 1, 2, 3, 16, 17, 18, 19, 4, 5, 6, 7, 20, 21, 22, 23]
        assert indices[-6:] == [12, 13, 14, 15, 28, 29]
    return indices


def test_resume():
    batch_size = 4
    from_beginning_indices = check_sampler_indices(shuffle=False, batch_size=batch_size, resume_step=0)
    check_sampler_indices(shuffle=True, batch_size=batch_size, resume_step=0)
    resume_step = 3
    resumed_indices = check_sampler_indices(shuffle=False, batch_size=batch_size, resume_step=resume_step)
    assert resumed_indices == from_beginning_indices[batch_size * resume_step:]


def test_get_indices_with_ratio():
    # Test setup
    ds = SimpleDataset(1, 10)  # Single dataset with 10 elements
    sampler = AncSampler(ds, batch_size=2, shuffle=False)
    base_indices = list(range(10))  # [0,1,2,3,4,5,6,7,8,9]
    
    # Test case 1: Integer ratio without shuffle
    result = sampler._get_indices_with_ratio(base_indices, 2.0)
    assert len(result) == 2, "Should have 2 copies of indices"
    assert result[0] == base_indices, "First copy should match original"
    assert result[1] == base_indices, "Second copy should match original"
    
    # Test case 2: Fractional ratio without shuffle
    result = sampler._get_indices_with_ratio(base_indices, 1.5)
    assert len(result) == 2, "Should have 2 lists (1 full + 1 partial)"
    assert result[0] == base_indices, "First copy should be complete"
    assert len(result[1]) == 5, "Second copy should be half length"
    assert result[1] == base_indices[:5], "Partial copy should be first half"
    
    # Test case 3: With chunk granularity
    sampler.chunk_granularity = 3
    result = sampler._get_indices_with_ratio(base_indices, 1.0)
    assert len(result) == 4, "Should split into 4 chunks of size 3 (last chunk size 1)"
    assert result[0] == [0,1,2], "First chunk"
    assert result[1] == [3,4,5], "Second chunk"
    assert result[2] == [6,7,8], "Third chunk"
    assert result[3] == [9], "Last chunk"

    # Test case 4: With shuffle
    sampler.shuffle = True
    sampler.chunk_granularity = -1
    random.seed(42)  # For reproducibility
    result = sampler._get_indices_with_ratio(base_indices, 1.0)
    assert len(result) == 1, "Should have 1 shuffled copy"
    assert len(result[0]) == len(base_indices), "Should contain all indices"
    assert result[0] != base_indices, "Should be shuffled"
    assert sorted(result[0]) == base_indices, "Should contain same elements"


if __name__ == "__main__":
    test_resume()
    test_get_indices_with_ratio()


import json
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os
def read_benchmark_results(file_path):
    with open(file_path, "r") as f:
        return json.load(f)

def analyze_benchmark_results_anc(file_path, base_result_dir):
    """
    Analyze benchmark results by creating a pandas DataFrame and generating summary statistics.
    
    Args:
        benchmark_results_list: List of benchmark result dictionaries
        base_result_dir: Directory to save analysis results
    """
    try:
        # Check if file_path is a directory or a single file
        all_results = []
        if os.path.isdir(file_path):
            # Process all JSON files in the directory
            print(f"Processing all JSON files in the directory: {file_path}")
            json_files = [f for f in os.listdir(file_path) if f.endswith('.json')]
            for json_file in json_files:
                full_path = os.path.join(file_path, json_file)
                file_results = read_benchmark_results(full_path)
                # Add filename as a column (without .json extension)
                file_name = os.path.basename(json_file).replace('.json', '')
                for result in file_results:
                    result['file_name'] = file_name
                all_results.extend(file_results)
        else:
            # Process a single file
            file_results = read_benchmark_results(file_path)
            file_name = os.path.basename(file_path).replace('.json', '')
            for result in file_results:
                result['file_name'] = file_name
            all_results.extend(file_results)
            
        # Convert list of dictionaries to DataFrame
        df = pd.DataFrame(all_results)

        # Recalculate request_throughput and output_throughput based on tensor_parallel_size
        if 'tensor_parallel_size' in df.columns and 'request_throughput' in df.columns:
            # Convert tensor_parallel_size to numeric type to avoid TypeError
            df['tensor_parallel_size'] = pd.to_numeric(df['tensor_parallel_size'])
            df['request_throughput'] = df['request_throughput'] * 8 / df['tensor_parallel_size']
        
        if 'tensor_parallel_size' in df.columns and 'output_throughput' in df.columns:
            # Ensure tensor_parallel_size is numeric
            if not pd.api.types.is_numeric_dtype(df['tensor_parallel_size']):
                df['tensor_parallel_size'] = pd.to_numeric(df['tensor_parallel_size'])
            df['output_throughput'] = df['output_throughput'] * 8 / df['tensor_parallel_size']

        grid_search_columns = ["enable_chunked_prefill",
                               "enable_prefix_caching",
                               "use-v2-block-manager",
                               "multi-step-stream-outputs",
                               "tensor_parallel_size",
                               "speculative-draft-tensor-parallel-size",
                               "num-speculative-tokens"]
        
        select_columns = ["request_throughput",
                          "output_throughput", 
                          "mean_output_token_len",
                          "mean_input_token_len",
                          "max_concurrency", 
                          "mean_ttfs_ms",
                          "std_ttfs_ms",
                          "p90_ttfs_ms",
                          "mean_tpot_ms",
                          "std_tpot_ms",
                          "p90_tpot_ms",
                          "mean_ttft_ms",
                          "std_ttft_ms",
                          "p90_ttft_ms",
                          "mean_e2el_ms",
                          "std_e2el_ms",
                          "p90_e2el_ms",
                          "mean_fs_token_len",
                          "std_fs_token_len",
                          "p90_fs_token_len",
         ]
        if 'file_name' in df.columns:
            select_columns = ["file_name"] + select_columns


        # Filter out columns that have only a single unique value
        # df = df[(df["enable_prefix_caching"] == "True") ]
        # df = df[(df["enable_prefix_caching"] == "True") & (df["max_concurrency"]<=8) & (df["speculative-draft-tensor-parallel-size"]=="4")]
        
        all_selected_columns = grid_search_columns + select_columns
        filtered_columns = []
        for col in all_selected_columns:
            if col in df.columns:
                if df[col].nunique() > 1:
                    filtered_columns.append(col)
        df = df[filtered_columns]
        # Rename columns
        renamed_grid_search_columns = {
            "enable_chunked_prefill": "chunkfill",
            "enable_prefix_caching": "precache",
            "tensor_parallel_size": "TP",
            "file_name": "file",
            "speculative-draft-tensor-parallel-size": "sp_tp",
            "num-speculative-tokens": "k"
        }
        df.rename(columns=renamed_grid_search_columns, inplace=True)
        # Calculate TTFS as TTFT + TPOT * output_token_len
        #recalculate_ttfs(df)
            
        df.to_csv(os.path.join(base_result_dir, "benchmark_results.csv"), index=False)
        print(f"Data frame of Benchmark results saved to {os.path.join(base_result_dir, 'benchmark_results.csv')}")
        
        # Create a figure with 4 subfigures
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(15, 10))
        
        # Create groups based on configuration parameters
        # Use grid_search_columns to create unique configuration identifiers
        df['config_id'] = df.apply(lambda row: '_'.join([f"{col}_{row[col]}" for col in renamed_grid_search_columns.values() if col in row]), axis=1)
        
        # Get unique configuration IDs
        config_ids = df['config_id'].unique()
        
        # Create groups based on configuration
        groups = []
        for config_id in config_ids:
            group_df = df[df['config_id'] == config_id]
            if not group_df.empty:
                groups.append((config_id, group_df))
        
        # Use different colors for different configurations
        colors = plt.cm.viridis(np.linspace(0, 1, len(groups)))
        
        # Define metrics to plot
        mean_metrics = [
            ('mean_ttft_ms', 'Mean TTFT (ms)'),
            ('mean_tpot_ms', 'Mean TPOT (ms)'),
            ('mean_ttfs_ms', 'Mean TTFS (ms)'),
            ('mean_e2el_ms', 'Mean E2EL (ms)')
        ]
        
        p90_metrics = [
            ('p90_ttft_ms', 'P90 TTFT (ms)'),
            ('p90_tpot_ms', 'P90 TPOT (ms)'),
            ('p90_ttfs_ms', 'P90 TTFS (ms)'),
            ('p90_e2el_ms', 'P90 E2EL (ms)')
        ]
        
        # Function to create plots with given x-axis
        def create_performance_plots(y_axis, y_label, filename, swap_axes=False):
            # default x-axis is max_concurrency
            fig, axes = plt.subplots(2, 4, figsize=(20, 10))
            
            # Flatten axes for easier iteration
            axes = axes.flatten()
            
            # Helper function to create a plot for a metric
            def plot_metric(ax, metric, metric_label, y_axis_name, groups_data, swap_axes=swap_axes):
                # Define different marker styles for different configurations
                markers = ['o', 's', '^', 'D', 'v', '<', '>', 'p', '*', 'h', 'H', '+', 'x', '|']
                
                for j, (config_id, group) in enumerate(groups_data):
                    # Use different marker styles based on configuration
                    marker_idx = j % len(markers)
                    marker_style = markers[marker_idx]
                    
                    if swap_axes:
                        ax.scatter(group[y_axis], group[metric],
                                  label=config_id, 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[y_axis], group[metric],
                               color=colors[j], linestyle='-')
                        ax.set_xlabel(y_label)
                        ax.set_ylabel(metric_label)
                        ax.set_title(f'{y_label} vs {metric_label}')
                    else:
                        ax.scatter(group[metric], group[y_axis],
                                  label=config_id, 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[metric], group[y_axis],
                               color=colors[j], linestyle='-')
                        ax.set_ylabel(y_label)
                        ax.set_xlabel(metric_label)
                        ax.set_title(f'{metric_label} vs {y_label}')
                        # ax.set_xlim(right=4000)
                        # ax.set_ylim(top=7)
                ax.grid(True)
                
                # Only add legend to the first graph (index 0)
                if ax == axes[0]:
                    ax.legend(fontsize='x-small', loc='upper right', frameon=True)
            
            # Create two sets of plots - one with default axes and one with swapped axes
            # Plot mean metrics (first row)
            for i, (metric, ylabel) in enumerate(mean_metrics):
                plot_metric(axes[i], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            # Plot p90 metrics (second row)
            for i, (metric, ylabel) in enumerate(p90_metrics):
                plot_metric(axes[i+4], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            plt.tight_layout()
            plt.savefig(os.path.join(base_result_dir, filename))
            plt.close()
        # Create plots with concurrency as x-axis
        create_performance_plots('max_concurrency', 'Concurrency', 'concurrency_analysis.png', swap_axes=True)
        
        # Create plots with request throughput as x-axis
        create_performance_plots('request_throughput', 'Request Throughput', 'throughput_analysis.png')
            
    except ImportError as e:
        print(f"Could not analyze results: {e}")
        print("To enable analysis, install pandas and matplotlib: pip install pandas matplotlib")
    except Exception as e:
        print(f"Error analyzing benchmark results: {e}")

def analyze_benchmark_results(file_path, base_result_dir):
    """
    Analyze benchmark results by creating a pandas DataFrame and generating summary statistics.
    
    Args:
        benchmark_results_list: List of benchmark result dictionaries
        base_result_dir: Directory to save analysis results
    """
    try:
        import pandas as pd
        import matplotlib.pyplot as plt

        benchmark_results_list = read_benchmark_results(file_path)
        # Convert list of dictionaries to DataFrame
        df = pd.DataFrame(benchmark_results_list)

        select_columns = ["request_throughput",
                          "output_throughput", 
                          "tensor_parallel_size", 
                          "mean_output_token_len",
                          "mean_input_token_len",
                          "max_concurrency", 
                          "mean_ttfs_ms",
                          "std_ttfs_ms",
                          "p90_ttfs_ms",
                          "mean_tpot_ms",
                          "std_tpot_ms",
                          "p90_tpot_ms",
                          "mean_ttft_ms",
                          "std_ttft_ms",
                          "p90_ttft_ms",
                          "mean_e2el_ms",
                          "std_e2el_ms",
                          "p90_e2el_ms"]
        # Calculate TTFS as TTFT + TPOT * output_token_len
        recalculate_ttfs(df, first_sentence_len=20)
            
        df = df[select_columns]

        df.to_csv(os.path.join(base_result_dir, "benchmark_results.csv"), index=False)
        print(f"Data frame of Benchmark results saved to {os.path.join(base_result_dir, 'benchmark_results.csv')}")
        
        # Create a figure with 4 subfigures
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(15, 10))
        
        # Group data by input token length and output token length
        distinct_input_lens = df['mean_input_token_len'].drop_duplicates().values
        distinct_output_lens = df['mean_output_token_len'].drop_duplicates().values
        
        # Create groups based on both input and output token lengths
        groups = []
        for input_len in distinct_input_lens:
            input_df = df[df['mean_input_token_len'] == input_len]
            for output_len in distinct_output_lens:
                group_df = input_df[input_df['mean_output_token_len'] == output_len]
                if not group_df.empty:
                    groups.append((input_len, output_len, group_df))
        
        # Use different colors for different combinations
        colors = plt.cm.viridis(np.linspace(0, 1, len(groups)))
        
        # Define metrics to plot
        mean_metrics = [
            ('mean_ttft_ms', 'Mean TTFT (ms)'),
            ('mean_tpot_ms', 'Mean TPOT (ms)'),
            ('mean_ttfs_ms', 'Mean TTFS (ms)'),
            ('mean_e2el_ms', 'Mean E2EL (ms)')
        ]
        
        p90_metrics = [
            ('p90_ttft_ms', 'P90 TTFT (ms)'),
            ('p90_tpot_ms', 'P90 TPOT (ms)'),
            ('p90_ttfs_ms', 'P90 TTFS (ms)'),
            ('p90_e2el_ms', 'P90 E2EL (ms)')
        ]
        
        # Function to create plots with given x-axis
        def create_performance_plots(y_axis, y_label, filename, swap_axes=False):
            fig, axes = plt.subplots(2, 4, figsize=(20, 10))
            
            # Flatten axes for easier iteration
            axes = axes.flatten()
            
            # Plot mean metrics (first row)
            # Helper function to create a plot for a metric
            def plot_metric(ax, metric, metric_label, y_axis_name, groups_data, swap_axes=swap_axes):
                # Define different marker styles for different output token lengths
                markers = ['o', 's', '^', 'D', 'v', '<', '>', 'p', '*', 'h', 'H', '+', 'x', '|']
                
                for j, (input_len, output_len, group) in enumerate(groups_data):
                    # Use different marker styles based on output token length
                    marker_idx = list(distinct_output_lens).index(output_len) % len(markers)
                    marker_style = markers[marker_idx]
                    
                    if swap_axes:
                        ax.scatter(group[y_axis], group[metric],
                                  label=f'In: {input_len}, Out: {output_len}', 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[y_axis], group[metric],
                               color=colors[j], linestyle='-')
                        ax.set_xlabel(y_label)
                        ax.set_ylabel(metric)
                        ax.set_title(f'{y_label} vs {metric}')
                    else:
                        ax.scatter(group[metric], group[y_axis],
                                  label=f'In: {input_len}, Out: {output_len}', 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[metric], group[y_axis],
                               color=colors[j], linestyle='-')
                        ax.set_ylabel(y_label)
                        ax.set_xlabel(metric)
                        ax.set_title(f'{metric} vs {y_label}')
                ax.grid(True)
                ax.legend()
            
            # Create two sets of plots - one with default axes and one with swapped axes
            # Plot mean metrics (first row) - default axes
            for i, (metric, ylabel) in enumerate(mean_metrics):
                plot_metric(axes[i], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            # Plot p90 metrics (second row) - default axes
            for i, (metric, ylabel) in enumerate(p90_metrics):
                plot_metric(axes[i+4], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            plt.tight_layout()
            plt.savefig(os.path.join(base_result_dir, filename))
            plt.close()
            
        # Create plots with concurrency as x-axis
        create_performance_plots('max_concurrency', 'Concurrency', 'concurrency_analysis.png', swap_axes=True)
        
        # Create plots with request throughput as x-axis
        create_performance_plots('request_throughput', 'Request Throughput', 'throughput_analysis.png')
            
    except ImportError as e:
        print(f"Could not analyze results: {e}")
        print("To enable analysis, install pandas and matplotlib: pip install pandas matplotlib")
    except Exception as e:
        print(f"Error analyzing benchmark results: {e}")

def recalculate_ttfs(df, first_sentence_len):
    if 'mean_ttfs_ms' in df.columns and 'mean_ttft_ms' in df.columns and 'mean_tpot_ms' in df.columns and 'mean_output_token_len' in df.columns:
        df['mean_ttfs_ms'] = df['mean_ttft_ms'] + df['mean_tpot_ms'] * first_sentence_len
            
    if 'p90_ttfs_ms' in df.columns and 'p90_ttft_ms' in df.columns and 'p90_tpot_ms' in df.columns and 'p90_output_token_len' in df.columns:
        df['p90_ttfs_ms'] = df['p90_ttft_ms'] + df['p90_tpot_ms'] * first_sentence_len
            
    if 'std_ttfs_ms' in df.columns and 'std_ttft_ms' in df.columns and 'std_tpot_ms' in df.columns and 'std_output_token_len' in df.columns:
            # This is an approximation for std of sum of dependent variables
        df['std_ttfs_ms'] = np.sqrt(df['std_ttft_ms']**2 + (df['std_tpot_ms'] * first_sentence_len)**2)
    # Save the raw DataFrame to CSV


import click
import os
from anc.cli.load_test.load_testing_operator import LoadTestingOperator
from anc.cli.load_test.backend_request_func import ASYNC_REQUEST_FUNCS
from anc.cli.load_test.data_analysis import analyze_benchmark_results, analyze_benchmark_results_anc
import argparse
import json
from .util import click_group

# there are two subcommands: run and plot
# run is used to start the server and send the benchmark requests, this will generate a json file with the results, by default it will also plot the results
# plot is used to plot the results from the json file. 

defaults = {
            'enable_chunked_prefill': "True",
            'enable_prefix_caching': "True",
            'use_v2_block_manager': "False",
            'gpu_memory_utilization': "0.95",
            'max-model-len': "8000",
            'multi_step_stream_outputs': "True",
            'tensor_parallel_size': "4",
            'num-speculative-tokens': "5",
            "speculative-draft-tensor-parallel-size": "1",
        }
def parse_arg_list(ctx, param, value):
    if value is None:
        return None
    try:
        return [int(x.strip()) for x in value.split(',')]
    except ValueError:
        raise click.BadParameter('values must be integers')

def parse_arg_float(ctx, param, value):
    if value is None:
        return None
    try:
        return [float(x.strip()) for x in value.split(',')]
    except ValueError:
        raise click.BadParameter('values must be integers')

def parse_arg_str(ctx, param, value):
    if value is None:
        return None
    return [str(x.strip()) for x in value.split(',')]

def validate_dataset_path(ctx, param, value):
    if ctx.params.get("dataset_name") == "anc" and not value:
        raise click.BadParameter("dataset-path is required when dataset-name is 'anc'.")
    return value

@click_group()
def loadtest():
    """Commands for load testing and benchmark analysis."""
    pass

@loadtest.command(name="run")
# Server configuration options
@click.option("--backend", default="vllm", type=click.Choice(list(ASYNC_REQUEST_FUNCS.keys())), help="Backend to use, for anc data set, we will use openai-chat backend")
@click.option("--model", type=str, required=True, help="Path of the model and tokenizer")
@click.option("--model-id", type=str, required=False, help="model name or id to use during load testing")
@click.option("--speculative-model", callback=parse_arg_str, required=False, help="path to draft model")
@click.option("--speculative-draft-tensor-parallel-size", callback=parse_arg_str, default= "1", type=str, required=False, help="Name of the model.")
@click.option("--num-speculative-tokens", callback=parse_arg_str, type=str, default="5",required=False, help="Number of speculative tokens.")
@click.option("--speculative-disable-by-batch-size", type=str, default="1000",required=False, help="turn off speculative decoding if the batch size is larger than the specified value")
@click.option("--host", type=str, default="localhost", help="Host address.")
@click.option("--port", type=int, default=8000, help="Port number.")
@click.option("--base-url", type=str, default=None, help="Server or API base url if not using http host and port.")
@click.option("--start_server_only", is_flag=True, help="Start the server only, do not send the benchmark requests. this is useful for debugging")
@click.option("--skip-server", is_flag=True, help="Skip server start. This is useful for debugging")
@click.option("--dry-run", is_flag=True, help="Dry run. This is useful for debugging")
@click.option("--reuse-server-each-concur", is_flag=True, help="reuse the server for each concurrency level. you can turn this off without prefix caching")
@click.option("--tensor-parallel-size", 
              callback=parse_arg_str,
              default=defaults["tensor_parallel_size"], 
              help="Tensor parallel size for model parallelism. Specify multiple values separated by commas (e.g., --tensor-parallel-size '1,2,4').")
@click.option("--enable-chunked-prefill", 
              callback=parse_arg_str,
              default=defaults["enable_chunked_prefill"], 
              help="Enable chunked prefill for vLLM. Specify multiple values separated by commas.")
@click.option("--enable-prefix-caching", 
              callback=parse_arg_str,
              default=defaults["enable_prefix_caching"], 
              help="Enable prefix caching for vLLM. Specify multiple values separated by commas.")
@click.option("--use-v2-block-manager", 
              callback=parse_arg_str,
              default=defaults["use_v2_block_manager"], 
              help="Use v2 block manager for vLLM. Specify multiple values separated by commas.")
@click.option("--gpu-memory-utilization", 
              default=defaults["gpu_memory_utilization"], 
              help="GPU memory utilization for vLLM. Specify multiple values separated by commas.")
@click.option("--max-model-len", 
              callback=parse_arg_str,
              default=defaults["max-model-len"], 
              help="Maximum model length for vLLM. Specify multiple values separated by commas.")
@click.option("--multi-step-stream-outputs", 
              callback=parse_arg_str,
              default=defaults["multi_step_stream_outputs"], 
              help="Enable multi-step stream outputs for vLLM. Specify multiple values separated by commas.")
# Load testing configuration options
@click.option("--endpoint", type=str, default="/v1/completions", help="API endpoint.")
@click.option("--dataset-name", type=click.Choice(["anc", "random", "hf"]), default="random", help="Name of the dataset to benchmark on.")
@click.option("--dataset-path", type=str, default=None, help="Path to the load testing dataset, required if using anc dataset.",callback=validate_dataset_path)
@click.option("--multi-turn", type=bool, default=False, help="Use multi-turn conversations, only used for anc dataset. if True, then we will send each prompt in multi-turn conversations in order. num_requests is the total number of conversations")
@click.option("--max-concurrency", 
              callback=parse_arg_list,
              help="Maximum number of concurrent requests. Specify multiple values separated by commas (e.g., --max-concurrency '1,4,8,16')")
@click.option("--tokenizer", type=str, help="Name or path of the tokenizer, if not using the default tokenizer.")
@click.option("--best-of", type=int, default=1, help="Generates `best_of` sequences per prompt and returns the best one.")
@click.option("--use-beam-search", is_flag=True, help="Use beam search.")
@click.option("--num-prompts", type=int, default=1000, help="Number of prompts to process.")
@click.option("--num-prompts-per-concurrency", type=int, default=None, help="Number of prompts for concurrency test. if specified, we will set num_prompts to num_prompts_per_concurrency * max_concurrency")
@click.option("--num-warmup-requests", type=int, default=0, help="Number of warmup requests to send before the actual load test.")
@click.option("--logprobs", type=int, default=None, help="Number of logprobs-per-token to compute & return as part of the request.")
@click.option("--request-rate", type=float, default=float("inf"), help="Number of requests per second.")
@click.option("--burstiness", type=float, default=1.0, help="Burstiness factor of the request generation.")
@click.option("--seed", type=int, default=None, required=False, help="if specified, we will set the random seed to the specified value. this insures same requests are sent for each concurrency")
@click.option("--trust-remote-code", is_flag=True, help="Trust remote code from huggingface.")
@click.option("--disable-tqdm", is_flag=True, help="Disable tqdm progress bar.")
@click.option("--profile", is_flag=True, help="Use Torch Profiler.")
@click.option("--save-result", is_flag=True, help="Save benchmark results to a json file.")
@click.option("--metadata", type=str, multiple=True, help="Key-value pairs for metadata of this run.")
@click.option("--result-dir", type=str, default=None, help="Directory to save benchmark json results.")
@click.option("--result-filename", type=str, default=None, help="Filename to save benchmark json results.")
@click.option("--ignore-eos", is_flag=True, help="Set ignore_eos flag when sending the benchmark request.")
@click.option("--percentile-metrics", type=str, default="input_token,output_token,fs_token,ttft,ttfs,tpot,itl,e2el", help="Comma-separated list of selected metrics to report percentiles.")
@click.option("--metric-percentiles", type=str, default="90", help="Comma-separated list of percentiles for selected metrics.")
@click.option("--goodput", type=str, multiple=True, help="Specify service level objectives for goodput as `KEY:VALUE` pairs.")

# Dataset-specific options
@click.option("--anc-output-len", type=int, default=None, help="Output length for each request. Overrides the output length from the anc dataset.")
@click.option("--random-input-len", 
              callback=parse_arg_list,
              default="1024", 
              help="Number of input tokens per request, used only for random sampling. Specify multiple values separated by commas.")
@click.option("--random-output-len", 
              callback=parse_arg_list,
              default="128", 
              help="Number of output tokens per request, used only for random sampling. Specify multiple values separated by commas.")
@click.option("--random-range-ratio", 
              default=1.0, 
              help="Range of sampled ratio of input/output length, used only for random sampling")
@click.option("--random-prefix-len", 
              callback=parse_arg_list,
              default="0", 
              help="Number of fixed prefix tokens before random context. Specify multiple values separated by commas.")
@click.option("--hf-subset", type=str, default=None, help="Subset of the HF dataset.")
@click.option("--hf-split", type=str, default=None, help="Split of the HF dataset.")
@click.option("--hf-output-len", type=int, default=None, help="Output length for each request. Overrides the output lengths from the sampled HF dataset.")
@click.option("--tokenizer_mode", type=click.Choice(['auto', 'slow', 'mistral']), default="auto", help="The tokenizer mode.")
def run_loadtest(**kwargs):
    """Run load testing with the specified parameters."""
    # Convert kwargs to argparse.Namespace
    args = argparse.Namespace(**kwargs)
    
    reset_result_output_argument(args)
    op = LoadTestingOperator()
    
    # Print key server configuration parameters for better debugging
    # Define all server parameters that can have multiple values
    grid_search_server_params = {
        'tensor_parallel_size': args.tensor_parallel_size,
        'enable_chunked_prefill': args.enable_chunked_prefill,
        'enable_prefix_caching': args.enable_prefix_caching,
        'use-v2-block-manager': args.use_v2_block_manager,
        'max-model-len': args.max_model_len,
        'multi_step_stream_outputs': args.multi_step_stream_outputs,
        # Add more parameters here as needed
    }
    if args.speculative_model:
        grid_search_server_params.update({
            'speculative-model': args.speculative_model,
            'speculative-draft-tensor-parallel-size': args.speculative_draft_tensor_parallel_size,
            'num-speculative-tokens': args.num_speculative_tokens,
        })
    # Generate all combinations of parameters
    import itertools
    param_names = list(grid_search_server_params.keys())
    param_values = [grid_search_server_params[name] for name in param_names]

    max_concurrency_list = args.max_concurrency
    random_prefix_len_list = args.random_prefix_len
    random_input_len_list = args.random_input_len
    random_output_len_list = args.random_output_len
    benchmark_results_list = []
    base_result_dir = args.result_dir if args.result_dir else "./"
    for values in itertools.product(*param_values):
        # Create a dictionary with the current combination of parameters
        param_dict = dict(zip(param_names, values))
        
        # Create the server configuration with fixed and variable parameters
        server_config = {
            'backend': args.backend,
            'model': args.model,
            'host': args.host,
            'port': args.port,
            'gpu_memory_utilization': args.gpu_memory_utilization,
            'speculative_disable_by_batchsize': args.speculative_disable_by_batch_size,
            **param_dict  # Unpack the current parameter combination
        }
        print(f'server_config: {server_config}')

        for key, value in server_config.items():
            print(f'  {key}: {value}')

        print(f'\n\n{"="*100}\n\n')
        set_result_dir(base_result_dir, args, param_dict)
       
       
        start_server(args, op, server_config)
        if args.dry_run:
            continue
        for i in range(len(max_concurrency_list)):
            max_concurrency = max_concurrency_list[i]
            # skip restarting server for the first concurrency run
            if i!=0 and not args.reuse_server_each_concur and not args.skip_server:
                op.stop_server()
                start_server(args, op, server_config)
            if args.num_prompts_per_concurrency:
                args.num_prompts = args.num_prompts_per_concurrency * max_concurrency
            if args.dataset_name == "random":
                for random_prefix_len in random_prefix_len_list:
                    for input_len in random_input_len_list:
                        for output_len in random_output_len_list:
                                args.max_concurrency = max_concurrency
                                args.random_prefix_len = random_prefix_len
                                args.random_input_len = input_len
                                args.random_output_len = output_len
                                args.result_filename = f"all_results.json"
                                benchmark_result = op.load_test(args)
                                client_config = {
                                    'max_concurrency': max_concurrency,
                                    'random_prefix_len': random_prefix_len,
                                    'random_input_len': input_len,
                                    'random_output_len': output_len,
                                }
                                # Combine server and client configs with benchmark results
                                result_json = {
                                    **server_config,
                                    **client_config,
                                    **benchmark_result
                                }
                                benchmark_results_list.append(result_json)
                                # Save all benchmark results to a JSON file
                                result_file_path = os.path.join(base_result_dir, args.result_filename)
                                with open(result_file_path, "w") as f:
                                    json.dump(benchmark_results_list, f, indent=2)
            else:
                client_config = {
                    'max_concurrency': max_concurrency,
                }
                args.max_concurrency = max_concurrency
                args.result_filename = f"all_results.json"
                benchmark_result = op.load_test(args)
                client_config = {
                    'max_concurrency': max_concurrency,
                }
                                # Combine server and client configs with benchmark results
                result_json = {
                                    **server_config,
                                    **client_config,
                                    **benchmark_result
                                }
                benchmark_results_list.append(result_json)
                                # Save all benchmark results to a JSON file
                result_file_path = os.path.join(base_result_dir, args.result_filename)
                with open(result_file_path, "w") as f:
                    json.dump(benchmark_results_list, f, indent=2)
        if not hasattr(args, 'skip_server') or not args.skip_server:
            op.stop_server()
    
    # After all tests are complete, analyze the results
    if args.result_filename:
        result_file_path = os.path.join(base_result_dir, args.result_filename)
        if os.path.exists(result_file_path):
            if args.dataset_name == "anc":
                analyze_benchmark_results_anc(result_file_path, base_result_dir)
            else:
                analyze_benchmark_results(result_file_path, base_result_dir)
            print(f"Analysis complete. Results saved to {base_result_dir}")

def start_server(args, op, server_config):
    print(f'Result directory: {args.result_dir}')
    if not hasattr(args, 'skip_server') or not args.skip_server:
        print(f'Starting server with configuration:')
        op.start_server(args.dry_run, args.result_dir, **server_config)
        if args.start_server_only:
            print(f'Server started. Exiting.')
            exit()
    else:
        print(f'Skipping server start as requested')

@loadtest.command(name="plot")
@click.argument("result_file", type=click.Path(exists=True))
@click.option("--output-dir", type=str, default=None, help="Directory to save analysis results")
@click.option("--dataset-name", type=click.Choice(["anc", "random", "hf"]), default="random", help="Name of the dataset used to generated the result file. default is random ")
def plot_results(result_file, output_dir, dataset_name):
    """Analyze and plot benchmark results from a previously saved JSON file."""
    print(f"result_file: {result_file}")
    print(f"output_dir: {output_dir}")
    print(f"dataset_name: {dataset_name}")
    if not output_dir:
        output_dir = os.path.dirname(result_file)
        if not output_dir:
            output_dir = "."
    
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    
    print(f"Analyzing benchmark results from {result_file}")
    if dataset_name == "anc":
        analyze_benchmark_results_anc(result_file, output_dir)
    else:
        analyze_benchmark_results(result_file, output_dir)
    print(f"Analysis complete. Results saved to {output_dir}")

def reset_result_output_argument(args):
    if args.dataset_name == "random":
        args.ignore_eos = True

    if args.result_filename and not args.result_dir:
        print(f'Warning: --result-filename is specified but --result-dir is not. Using the current directory {os.getcwd()} as the result directory.')
        args.result_dir = os.path.dirname("")
        
    if args.result_dir and not os.path.exists(args.result_dir):
        print(f'Warning: --result-dir {args.result_dir} does not exist. Creating it.')
        os.makedirs(args.result_dir)
    
    if args.result_dir or args.result_filename:
        args.save_result = True

    if args.dataset_name == "anc":
        args.backend = "openai-chat"
        args.endpoint = "/v1/chat/completions"

def set_result_dir(base_result_dir, args, param_dict):
    # Add this line to use the base_result_dir from args
    
    dir_components = []
    # Define abbreviations and default values for parameters
    abbreviations = {
        'tensor_parallel_size': 'tp',
        'enable_chunked_prefill': 'chunk',
        'enable_prefix_caching': 'prefix',
        'use-v2-block-manager': 'v2block',
        'multi_step_stream_outputs': 'multistep',
    }
    
    # Default values to skip
    for key, value in param_dict.items():
        # Skip if using default value
        if key in defaults and value == defaults[key]:
            continue
            
        # Use abbreviation if available
        param_name = abbreviations.get(key, key.replace('-', '_'))
        dir_components.append(f'{param_name}_{value}')
    
    config_dir_name = '_'.join(dir_components)
    args.result_dir = os.path.join(base_result_dir, config_dir_name)
    if not os.path.exists(args.result_dir):
        os.makedirs(args.result_dir)
